'use strict';

/*******************************************************************/
/* Test Cases  */

	/* [Permissions checked!] */

App.Cases = new function () {
	var self = this;

	// Fields for keeping the current state (drag & drop, etc.)
	self.dragged_ids = null;
	self.droppables = null;
	self.formCopy = null;
	self.drop_data = null;

	self.applyActions = function () {
		self._applyHover();
	};

	self._applyHover = function () {
		// Late binding for drag & drop for cases. We delay setting up
		// the draggable/droppable until we hover over the drag action.

		$(document).on({
			mouseenter: function mouseenter() {
				if (!App.DragDrop.isDragging()) {
					self._applyDraggable($(this));
				}
			} }, 'td.drag');
	};

	self.applyResponsive = function () {
		App.Responsive.register('#content', 950, function (is_below) {
			var buttons = $('#content-header .button-responsive');
			if (is_below) {
				buttons.addClass('button-notext');
			} else {
				buttons.removeClass('button-notext');
			}
		});
	};

	//---------------------------------------------------------------
	// DRAG & DROP
	//---------------------------------------------------------------

	self._applyDraggable = function (column) {
		$('.caseDraggable', column).draggable({
			addClasses: false,
			cursor: 'move',
			containment: 'document',
			cursorAt: { top: 10, left: 5 },
			scroll: true,
			scrollSensitivity: 50,
			scrollSpeed: 40,
			scope: 'cases',
			start: self._dragStart,
			stop: self._dragStop,
			helper: self._dragHelper
		});
	};

	self._applyDroppable = function () {
		self.droppables = $('tr.caseDroppable, #groupTree a.caseDroppable');

		self.droppables.droppable({
			addClasses: false,
			tolerance: 'pointer',
			scope: 'cases',
			hoverClass: 'jstree-dropping',
			over: self._dropEnter,
			out: self._dropLeave,
			drop: self._drop
		});
	};

	self._destroyDroppable = function () {
		if (self.droppables) {
			self.droppables.droppable('destroy');
		}
	};

	self._dragHelper = function (event) {
		self._dragInit(this, event);
		return self._getDragHelper(this, event);
	};

	self._dragInit = function (el, event) {
		// Get and remember dragged case IDs for later
		self.dragged_ids = self._getDragged(el);
	};

	self._getDragged = function (el) {
		// In case the dragged row is selected (its checkbox is checked)
		// we return all selected rows, otherwise only the dragged case.

		var tr = $(el).closest('tr');
		var checkbox = $('input.selectionCheckbox', tr);

		if (!checkbox.is(':checked')) {
			return [tr.attr('rel')];
		} else {
			return App.Tables.getSelected($('#groups'));
		}
	};

	self._getDragHelper = function (el, event) {
		var tr = $(el).closest('tr');

		var case_ids = self.dragged_ids;
		if (case_ids.length > 1) {
			var title = '<strong>' + case_ids.length + '</strong> test cases';
		} else {
			var title = $('.title', tr).html();
		}

		var dragGrayUrl = Consts.resourceBaseUrl + 'images/icons/dragGray.png';

		var content = '<img src="' + dragGrayUrl + '" /> ' + title;

		// The section helper is hidden by default and will be shown by
		// the _dragStart handler. This is a workaround for a bug/mis-
		// behavior in Google Chrome which changes the column width of
		// the drag handle (not sure why).

		var helper = '<div id="caseHelper" class="hidden dragdrop-helper">' + content + '</div>';

		return $(helper).get(0);
	};

	self._dragStart = function (event, ui) {
		if (self.drop_data) {
			// Cancel any previous open drag & drop operation.
			self._dropCancel();
		}

		App.DragDrop.start();
		$(ui.helper).show(); // See getDragHelper

		// Defer applying droppable until we start dragging.
		self._applyDroppable();
	};

	self._dragStop = function (event, ui) {
		// And destroy droppables again when drag is finished.
		self._destroyDroppable();
		App.DragDrop.stop();
	};

	self._dropShowBorder = function (el) {
		var rows = $('td, th', $(el));
		rows.addClass('row-dragged-show').removeClass('row-dragged-hide');
	};

	self._dropHideBorder = function (el) {
		var row = $(el);
		$('th', row).addClass('row-dragged-hide').removeClass('row-dragged-show');
		$('td', row).addClass('row-dragged-hide').removeClass('row-dragged-show');
	};

	self._dropShowHighlight = function (section_id) {
		var node = $('#node-' + section_id);
		$('> a', node).addClass('jstree-highlighted');
	};

	self._dropHideHighlight = function (section_id) {
		var node = $('#node-' + section_id);
		$('> a', node).removeClass('jstree-highlighted');
	};

	self._dropEnter = function (event, ui) {
		self._dropShowBorder(this);
	};

	self._dropLeave = function (event, ui) {
		self._dropHideBorder(this);
	};

	self._drop = function (event, ui) {
		var target = $(this);

		// Get and remember the drop data (e.g. the droppable, section
		// etc.).
		if (this.tagName == 'A') // Sidebar tree
			{
				var section_id = target.closest('li').attr('rel');

				self.drop_data = {
					drop_to_tree: true,
					droppable: this,
					section_id: section_id,
					case_ids: self.dragged_ids
				};

				// Keep the hover highlight visible/active for the section.
				self._dropShowHighlight(section_id);
			} else {
			var after_id = target.attr('rel'); // Undefined if header row
			var section_id = target.closest('.grid-container').attr('rel');

			self.drop_data = {
				drop_to_tree: false,
				droppable: this,
				after_id: after_id,
				section_id: section_id,
				case_ids: self.dragged_ids
			};
		}

		var e = event.originalEvent;
		var show_menu = false;

		// If the user holds the shift key when dropping the case, we
		// execute the copy directly. The same is true for ctrl/cmd +
		// move. Otherwise, we show a dropdown menu.
		if (e.shiftKey) {
			self.dropCopy();
		} else if (e.ctrlKey || e.metaKey) {
			if (self._dropCanMove(self.drop_data)) {
				self.dropMove();
			} else {
				show_menu = true;
			}
		} else {
			show_menu = true;
		}

		if (show_menu) {
			self._dropShowMenu(e, self.drop_data);
		}
	};

	self._dropShowMenu = function (e, drop_data) {
		if (self._dropCanMove(drop_data)) {
			$('#casesDndMoveDisabled').hide();
			$('#casesDndMove').show();
		} else {
			$('#casesDndMove').hide();
			$('#casesDndMoveDisabled').show();
		}

		App.Dropdowns.show('#casesDndDropdown', e.pageX, e.pageY);
	};

	self._dropCanMove = function (drop_data) {
		var droppable = drop_data.droppable;

		if (droppable.tagName == 'A') {
			// Dropping on a sidebar section is always okay.
			return true;
		}

		var row = $(droppable);
		if (row.hasClass('header')) {
			// Dropping on a grid header is always okay.
			return true;
		}

		// In all other cases, we need to check if the move operation
		// is allowed. Simply said, we do not allow to drop rows into
		// dragged rows.
		var lookup = {};

		var case_ids = self.dragged_ids;
		for (var i = 0; i < case_ids.length; i++) {
			lookup[case_ids[i]] = true;
		}

		var id = row.attr('rel');
		return !lookup[id];
	};

	self._dropHideMenu = function () {
		App.Dropdowns.hide('#casesDndDropdown');
	};

	self._dropStart = function (el) {
		// Add a progress-indicating row to the table right below the
		// actual droppable.
		var progressInlineUrl = Consts.resourceBaseUrl + 'images/animations/progressInline.gif';

		var column_count = $('td, th', el).length;
		var tr = '<tr>' + '<td colspan="' + column_count + '">' + '<img src="' + progressInlineUrl + '" />' + '</td>' + '</tr>';

		$(el).after(tr);
	};

	self._dropStop = function (el) {
		self._dropHideBorder(el);
		self._dropClearCheckboxes();
		$(el).next().remove(); // Remove progress row again
		self.drop_data = null;
	};

	self._dropClearCheckboxes = function () {
		// Disable all possible checkboxes, not just the dragged row(s)
		var grids = $('#groups .grid-container');

		$.each(grids, function (index, value) {
			App.Tables.setCheckboxes($(value), false);
		});

		self._disableMassActions();
	};

	self.dropCancel = function () {
		self._dropCancel();
	};

	self._dropCancel = function () {
		self._dropHideMenu();
		if (self.drop_data.drop_to_tree) {
			self._dropHideHighlight(self.drop_data.section_id);
		} else {
			self._dropHideBorder(self.drop_data.droppable);
		}

		self.drop_data = null;
	};

	self.dropCopy = function () {
		self._dropHideMenu();

		// Call the actual copy action, either copy to section (tree
		// in the sidebar) or after a table row.
		if (self.drop_data.drop_to_tree) {
			self._dropCopyToSection(self.drop_data);
		} else {
			self._dropCopy(self.drop_data);
		}
	};

	self._dropCopy = function (o) {
		self._dropStart(o.droppable);

		App.Ajax.call({
			target: '/cases/ajax_copy',

			arguments: {
				case_ids: o.case_ids,
				section_id: o.section_id,
				after_id: o.after_id,
				columns: App.Tables.columns_for_user
			},

			success: function success(html) {
				self._dropStop(o.droppable);

				var grid = $('#grid-' + o.section_id);

				if (o.after_id) {
					$('#row-' + o.after_id).after(html);
				} else {
					$('tr.header', grid).after(html);
				}

				App.Suites.onCasesAdded();
				App.Suites.refreshGroupCount(o.section_id);
			},

			error: function error(data) {
				self._dropStop(o.droppable);
				App.Ajax.handleError(data);
			}
		});
	};

	self._dropCopyToSection = function (o) {
		var node = $('#node-' + o.section_id);
		var a = node.children('a');

		a.addClass('jstree-loading');

		App.Ajax.call({
			target: '/cases/ajax_copy_to_section_end',

			arguments: {
				case_ids: o.case_ids,
				section_id: o.section_id,
				columns: App.Tables.columns_for_user
			},

			success: function success(html) {
				a.removeClass('jstree-loading jstree-highlighted');

				var grid = $('#grid-' + o.section_id);
				if (grid.length) {
					$(html).appendTo(grid);
				}

				self._dropClearCheckboxes();
				App.Suites.onCasesAdded();
				App.Suites.refreshGroupCount(o.section_id);
			},

			error: function error(data) {
				a.removeClass('jstree-loading');
				App.Ajax.handleError(data);
			}
		});
	};

	self.dropMove = function () {
		self._dropHideMenu();

		// Call the actual move action, either move to section (tree
		// in the sidebar) or after a table row.		
		if (self.drop_data.drop_to_tree) {
			self._dropMoveToSection(self.drop_data);
		} else {
			self._dropMove(self.drop_data);
		}
	};

	self._dropMove = function (o) {
		self._dropStart(o.droppable);

		App.Ajax.call({
			target: '/cases/ajax_move',

			arguments: {
				case_ids: o.case_ids,
				section_id: o.section_id,
				after_id: o.after_id
			},

			success: function success(data) {
				self._dropStop(o.droppable);

				var section = '#grid-' + o.section_id;
				var cases = self._dropGetRows(o.case_ids);

				if (o.after_id) {
					cases.insertAfter($('#row-' + o.after_id));
				} else {
					cases.insertAfter($('tr.header', section));
				}

				App.Suites.refreshGroupCounts();
			},

			error: function error(data) {
				self._dropStop(o.droppable);
				App.Ajax.handleError(data);
			}
		});
	};

	self._dropMoveToSection = function (o) {
		var node = $('#node-' + o.section_id);
		var a = node.children('a');

		a.addClass('jstree-loading');

		App.Ajax.call({
			target: '/cases/ajax_move_to_section_end',

			arguments: {
				case_ids: o.case_ids,
				section_id: o.section_id
			},

			success: function success(data) {
				a.removeClass('jstree-loading jstree-highlighted');
				var grid = $('#grid-' + o.section_id);
				var cases = self._dropGetRows(o.case_ids);

				if (grid.length) {
					cases.appendTo(grid);
				} else {
					cases.remove();
				}

				self._dropClearCheckboxes();
				App.Suites.refreshGroupCounts();
			},

			error: function error(data) {
				a.removeClass('jstree-loading');
				App.Ajax.handleError(data);
			}
		});
	};

	self._dropGetRows = function (case_ids) {
		var cases = false;

		for (var i = 0; i < case_ids.length; i++) {
			var row = $('#row-' + case_ids[i]);

			if (cases) {
				cases = cases.add(row);
			} else {
				cases = row;
			}
		}

		return cases;
	};
}();

//---------------------------------------------------------------------
// ATTACHMENTS
//---------------------------------------------------------------------

App.Cases.Attachments = new function () {
	var self = this;

	self.init = function (selector, project_id, case_id) {
		var target = $(selector);

		var dropzone = App.Dropzone.applyDrop(selector, {
			url: App.Page.formatUri('{0}/attachments/ajax_add_for_case/{1}/{2}', Consts.ajaxBaseUrl, project_id, case_id ? case_id : ''),

			dict: {
				drop: lang('attachments_drop')
			},

			start: function start() {
				// Hide volatile dropzones if we are dropping to a
				// persistent dropzone.
				if (target.is('.dz-persistent')) {
					App.Dropzone.hide();
				}
			},

			stop: function stop() {
				App.Dropzone.hide(); // Hide all volatile dropzones
			},

			success: function success(file, data) {
				self._addId(data.id);
				self._addRow(data.code);
				dropzone.removeFile(file);
			}
		});

		return dropzone;
	};

	self._addRow = function (html) {
		var row = $(html);
		$('#noCaseAttachments').hide();
		row.appendTo($('#caseAttachmentList')).show();
	};

	self._addId = function (id) {
		var attachments = $('#attachments');
		var value = attachments.val();
		attachments.val(value ? value + ',' + id : id);
	};

	self.remove = function (attachment_id) {
		App.Ajax.call({
			target: '/attachments/ajax_delete',

			arguments: {
				attachment_id: attachment_id
			},

			success: function success(data) {
				$('#attachment-' + attachment_id).remove();
				if ($('#caseAttachmentList div').length == 0) {
					$('#noCaseAttachments').show();
				}
			},

			error: function error(data) {
				App.Ajax.handleError(data);
			}
		});
	};
}();

App.Cases.Edit = new function () {
	var self = this;

	// Fields for keeping the current state (filters, case list, etc.)
	self.suite_id = null;
	self.scope = null;
	self.display = null;
	self.group_only = null;
	self.group_by = null;
	self.group_id = null;
	self.case_ids = null;
	self.filters = null;
	self.columns = null;

	self._createFilter = function (e) {
		return new App.Suites.Filter({
			event: e,
			suite_id: self.suite_id,
			filters: self.filters,
			save_filters: false,

			changed: function changed(filters) {
				self.filters = filters;
				self._filterReload();
			}
		});
	};

	self.filter = function (e) {
		var filter = self._createFilter(e);
		filter.open();
	};

	self.filterReset = function () {
		var filter = self._createFilter();
		filter.reset();
	};

	self._filterReload = function () {
		App.Ajax.call({
			target: '/cases/ajax_render_edit_all_filter_match',
			blockUI: false,

			// Submit all input fields of the form along with the suite
			// ID this is about.
			arguments: {
				suite_id: self.suite_id,
				display: self.display,
				group_only: self.group_only,
				group_by: self.group_by,
				group_id: self.group_id,
				filters: self.filters
			},

			success: function success(html) {
				$('#filterByMatch').html(html);
			},

			error: function error(data) {
				// Ignore error
			}
		});
	};

	self.includeChanged = function () {
		var control_id = $(this).attr('rel');
		var control = $('#' + control_id);

		// We differentiate between standard and special controls
		// (multi/steps). For the standard control, we simply toggle
		// the placeholder attribute. For the special controls, we
		// toggle between the special 'various' message and the actual
		// control container.

		if (control.is('.multi') || control.is('.steps')) {
			var various = $('#' + control_id + '_various');
			if (various.length) {
				var container = $('#' + control_id + '_container');
				if (this.checked) {
					various.hide();
					container.show();
				} else {
					container.hide();
					various.show();
				}
			}
		} else {
			if (this.checked) {
				control.data('placeholder', control.attr('placeholder'));
				control.removeAttr('placeholder');
				control.focus();
			} else {
				control.attr('placeholder', control.data('placeholder'));
			}
		}
	};

	self.controlChanged = function () {
		var control = $(this);
		var checkbox = $('#' + control.attr('id') + '_included');
		checkbox.attr('checked', true).change();
	};

	self.confirm = function (suite_id, submit) {
		App.Validation.hideErrors();
		var case_count = 0;

		// Submit action
		$('#confirmDiffForm').unbind('submit');
		$('#confirmDiffForm').submit(function (e) {
			// Depending on the amount of test cases to update, we also
			// display an additional confirmation dialog.
			if (case_count >= 50) {
				if (App.Dialogs.remove('l:cases_edit_many_diff_confirm_dialog', $('#confirmDiffSubmit').html(), null, // No extra
				null, // No init callback
				function () {
					App.Dialogs.closeTop();
					submit();
				})) ;
			} else {
				App.Dialogs.closeTop();
				submit();
			}

			return false;
		});

		$('#accept').addClass('button-busy');

		App.Ajax.call({
			target: '/cases/ajax_render_edit_many_diff',

			// Submit all input fields of the form along with the suite
			// ID this is about.
			arguments: $.extend({
				suite_id: suite_id,
				scope: self.scope,
				case_ids: self.getCases(),
				display: self.display,
				group_only: self.group_only,
				group_by: self.group_by,
				group_id: self.group_id,
				filters: self.filters
			}, self._getInput()),

			stop: function stop() {
				$('#accept').removeClass('button-busy');
			},

			success: function success(data) {
				case_count = data.case_count; // See top of function

				if (data.show_diff) {
					$('#diff').html(data.code);
					$('#confirmDiffSubmit').text(data.message);

					App.Dialogs.open({
						selector: '#confirmDiffDialog'
					});
				} else {
					// Do not display the diff dialog, submit directly.
					submit();
				}
			},

			error: function error(data) {
				App.Ajax.handleError(data);
			}
		});
	};

	self._getInput = function () {
		// If there's a step field, make sure to synchronize the steps
		// with the form field. The control would usually do this as
		// well on form submit. But with the new template field, it may
		// be loaded dynamically after the initial form load and confirm
		// may be called before the field can update the data.

		var steps = $('#form .custom.steps');
		if (steps.length) {
			App.Cases.updateSteps(steps.attr('id'));
		}

		values = {};

		$('#form :input').each(function () {
			var t = $(this);
			if (t.attr('type') == 'checkbox') {
				values[t.attr('name')] = t.prop('checked');
			} else {
				values[this.name] = $(this).val();
			}
		});

		return values;
	};

	self.setCases = function (case_ids) {
		self.case_ids = {};
		if (case_ids) {
			$.each(case_ids, function (ix, v) {
				self.case_ids[v] = true;
			});
		}
	};

	self.getCases = function () {
		return self.case_ids ? Object.keys(self.case_ids) : [];
	};

	self.removeCase = function (case_id) {
		var row = $('#row-' + case_id);
		if (row.next().length == 0) {
			row.prev().find('th, td').addClass('noborder');
		}

		row.remove();
		delete self.case_ids[case_id];
	};

	self.loadCases = function (offset) {
		$('#paginationBusy').show();

		App.Ajax.call({
			target: '/cases/ajax_render_edit_many_cases',

			arguments: {
				suite_id: self.suite_id,
				case_ids: self.getCases(),
				columns: self.columns,
				offset: offset
			},

			stop: function stop() {
				$('#paginationBusy').show();
			},

			success: function success(html) {
				$('#cases').html(html);
			},

			error: function error(data) {
				App.Ajax.handleError(data);
			}
		});
	};
}();

//---------------------------------------------------------------------
// TESTS/RESULTS
//---------------------------------------------------------------------

App.Cases.showTests = function () {
	$('#activitiesLink').removeClass('text-active');
	$('#tests').show();
	$('#testsLink').addClass('text-active');
	$('#activities').hide();
};

App.Cases.loadTests = function (case_id) {
	$('#showTests .showAll').hide();
	$('#showTests .busy').show();

	App.Ajax.call({
		target: '/cases/ajax_get_tests',

		arguments: {
			case_id: case_id
		},

		success: function success(html) {
			$('#showTests .busy').hide();
			$('#tests').html(html);
		},

		error: function error(data) {
			$('#showTests .busy').hide();
			App.Ajax.handleError(data);
		}
	});
};

App.Cases.showActivities = function () {
	$('#testsLink').removeClass('text-active');
	$('#activities').show();
	$('#activitiesLink').addClass('text-active');
	$('#tests').hide();
};

App.Cases.loadActivities = function (case_id) {
	$('#showActivities .showAll').hide();
	$('#showActivities .busy').show();

	App.Ajax.call({
		target: '/cases/ajax_get_activities',

		arguments: {
			case_id: case_id
		},

		success: function success(html) {
			$('#showActivities .busy').hide();
			$('#activities').html(html);
		},

		error: function error(data) {
			$('#showActivities .busy').hide();
			App.Ajax.handleError(data);
		}
	});
};

//---------------------------------------------------------------------
// HISTORY
//---------------------------------------------------------------------

App.Cases.showPrevious = function (id) {
	$('#previous-' + id).show();
	$('#previousLink-' + id).hide();
};

App.Cases.loadHistory = function (case_id) {
	$('#showHistory .showAll').hide();
	$('#showHistory .busy').show();

	App.Ajax.call({
		target: '/cases/ajax_get_history',

		arguments: {
			case_id: case_id
		},

		success: function success(html) {
			$('#showHistory .busy').hide();
			$('#history').html(html);
		},

		error: function error(data) {
			$('#showHistory .busy').hide();
			App.Ajax.handleError(data);
		}
	});
};

//---------------------------------------------------------------------
// DEFECTS
//---------------------------------------------------------------------

App.Cases.loadDefects = function (case_id) {
	$('#showDefects .showAll').hide();
	$('#showDefects .busy').show();

	App.Ajax.call({
		target: '/cases/ajax_render_defects',

		arguments: {
			case_id: case_id
		},

		stop: function stop() {
			$('#showDefects .busy').hide();
		},

		success: function success(html) {
			$('#defects').html(html);
		},

		error: function error(data) {
			App.Ajax.handleError(data);
		}
	});
};

//---------------------------------------------------------------------
// ADD/EDIT/DELETE
//---------------------------------------------------------------------

App.Cases.editTitleDialog = function (o) {
	App.Validation.hideErrors();

	// Initialize the dialog
	$('#editCaseTitle').val(o.title);
	$('#editCaseForm').unbind('submit');
	$('#editCaseSubmit').removeClass('button-busy');

	$('#editCaseForm').submit(function (e) {
		App.Validation.hideErrors();
		var title = $.trim($('#editCaseTitle').val());

		// Signal busy
		$('#editCaseSubmit').addClass('button-busy');

		o.submit(title);
		return false;
	});

	// Also make sure to accept Enter for submitting the form
	$('#editCaseTitle').unbind('keydown').bind('keydown', function (e) {
		if (e.keyCode == App.keyEnter) {
			$('#editCaseForm').submit();
			return false;
		}
	});

	App.Dialogs.open({
		selector: '#editCaseDialog',
		focusedControl: '#editCaseTitle',
		selectedControl: '#editCaseTitle',
		titleSelector: o.titleSelector
	});
	App.DirtyChecker.init("editCaseSubmit", "editCaseForm");
};

App.Cases.editTitle = function (case_id) {
	// Get the related case title
	var selector = '#row-' + case_id + ' .title';
	var title = $(selector).text();

	$('.editCaseEdit').show();
	$('.editCaseAdd').hide();

	// Show the dialog
	App.Cases.editTitleDialog({
		title: title,
		titleSelector: '.dialogTitleEdit',
		submit: function submit(title) {
			// Rename the section on the server
			App.Ajax.call({
				target: '/cases/ajax_edit_title',

				arguments: {
					case_id: case_id,
					title: title
				},

				success: function success(data) {
					$(selector).text(title);
					App.Effects.add(selector);
					App.Dialogs.close('#editCaseDialog');
				},

				error: function error(data) {
					$('#editCaseSubmit').removeClass('button-busy');
					App.Ajax.handleError(data, '#editCaseErrors');
				}
			});
		}
	});
};

App.Cases.callEditCase = function (case_id) {
	App.Ajax.call({
		target: '/cases/ajax_edit',
		arguments: App.Cases.serializeForm({case_id: case_id}),
		success: function success(data) {
			App.Cases.renderQpaneAgain(case_id);
			self.qpane.removeClass("edit-mode");
		},
		error: function error(data) {
			App.Ajax.handleError(data);
		}
	})
}

App.Cases.updateFromQpane = function (case_id) {

	var formData = App.Cases.serializeForm({case_id: case_id});

	App.Cases.callAjaxCheckCaseInDynamicFilters(case_id, formData).done(function(resp) {
		resp = JSON.parse(resp);
			if (resp.is_added) {
				App.Dialogs.confirm(resp.info, function() {
					App.Cases.callEditCase(case_id)
				}, function() {}, '#confirmDialogOkCancel');
			} else {
				App.Cases.callEditCase(case_id)
			}
	}).fail(function(resp) {
		App.Ajax.handleError(resp);
	})
};

App.Cases.renderQpaneAgain = function (test_id) {
	App.Ajax.call({
		target: '/cases/ajax_render_qpane',
		reflow: true, // For responsive qpane
		arguments: {
			case_id: test_id
		},
		success: function success(html) {
			$('#qpane-content').html(html);
		},
		error: function error(data) {
			App.Ajax.handleError(data);
		}
	});
};

App.Cases.resetFormBody = function () {
	if (self.formCopy != null) {
		$("#qpane form").html(self.formCopy);
	}
};

App.Cases.editMode = function (test_id) {
	self.qpane = $("#qpane");
	self.formCopy = self.qpane.find('form').html();

	if (self.qpane.hasClass("edit-mode")) {
		App.Cases.updateSteps('custom_steps_separated');
		App.Cases.updateFromQpane(test_id);

		return;
	}

	self.qpane.addClass("edit-mode");
	$('#cancelButton').show();
	$('#closeButton').hide();
};

App.Cases.cancelEditMode = function()
{
	self.qpane = $("#qpane");
	self.formCopy = self.qpane.find('form').html();

	if (self.qpane.hasClass("edit-mode")) {
		self.qpane.removeClass("edit-mode");
		$('#cancelButton').hide();
		$('#closeButton').show();
		App.Cases.resetFormBody();

		return;
	}
}

App.Cases.add = function (section_id) {
	App.Validation.hideErrors();

	// Make sure that all other inline forms are hidden
	$('.inlineSectionActions').show();
	$('.inlineSectionAddCase').hide();

	// Get the element names
	var actionsElement = '#inlineSectionActions-' + section_id;
	var addElement = '#inlineSectionAddCase-' + section_id;

	// Prepare the form
	$(addElement + ' .addForm').unbind('submit');
	$(addElement + ' .cancel').unbind('click');
	$(addElement + ' .title').unbind('keydown');
	$(addElement + ' .title').val('');

	// Prepare the functionality to hide/cancel the form
	var hideForm = function hideForm() {
		$(actionsElement).show();
		$(addElement).hide();
	};

	// Mark the form as busy / none-busy
	var signalBusy = function signalBusy(busy) {
		if (busy) {
			$(addElement + ' .buttons').hide();
			$(addElement + ' .busy').show();
		} else {
			$(addElement + ' .buttons').show();
			$(addElement + ' .busy').hide();
		}
	};

	// Make sure that the user can cancel the inline form with escape
	$(addElement + ' .title').keydown(function (e) {
		if (e.keyCode == App.keyEscape) {
			hideForm();
		}
	});

	// Assign the cancel functionality to the cancel image link
	$(addElement + ' .cancel').click(hideForm);

	var form = $(addElement + ' .addForm');

	// Specify what should happen when the user submits the case
	$(addElement + ' .addForm').submit(function () {
		$(addElement + ' .submit').blur();
		App.Validation.hideErrors();

		// Get and validate the case title
		var caseTitle = $.trim($(addElement + ' .title').val());

		if (!caseTitle) {
			$(addElement + ' .errorPanel').show();
			$(addElement + ' .requiredMessage').show();
			return false;
		}

		signalBusy(true);

		App.Ajax.call({
			target: 'cases/ajax_add',

			arguments: {
				section_id: section_id,
				title: caseTitle,
				columns: App.Tables.columns_for_user
			},

			success: function success(html) {
				$(addElement + ' .title').val('');
				$(addElement + ' .title').focus();
				signalBusy(false);

				var grid = $('#grid-' + section_id);
				var row = $(html);
				row.appendTo(grid);
				App.Effects.add(row);

				App.Suites.onCasesAdded();
				App.Suites.refreshGroupCount(section_id);

				if (App.Users.hasGoals()) {
					App.Users.reloadGoals();
				}
			},

			error: function error(data) {
				signalBusy(false);
				App.Ajax.handleError(data);
			}
		});

		return false;
	});

	$(actionsElement).hide();
	$(addElement).show();
	$(addElement + ' .title').focus();
};

App.Cases.remove = function (case_id) {
	var row = $('#row-' + case_id);
	$('.deleteLink', row).hide();
	$('.deleteBusy', row).show();

	var section_id = row.closest('.grid-container').attr('rel');

	App.Ajax.call({
		target: 'cases/ajax_delete',

		arguments: {
			case_id: case_id
		},

		stop: function stop() {
			$('.deleteBusy', row).hide();
			$('.deleteLink', row).show();
		},

		success: function success(data) {
			$('#row-' + case_id).remove();
			App.Suites.onCasesDeleted();
			App.Suites.refreshGroupCount(section_id);
			App.Cases._updateMassActions(); // May be needed as well
		},

		error: function error(data) {
			App.Ajax.handleError(data);
		}
	});
};

App.Cases.removeMany = function (suite_id) {
	var case_ids = App.Tables.getSelected($('#groups'));
	$('#deleteCases').addClass('button-busy');

	App.Ajax.call({
		target: 'cases/ajax_delete_many',

		arguments: {
			suite_id: suite_id,
			case_ids: case_ids
		},

		success: function success(data) {
			$('#deleteCases').removeClass('button-busy');
			App.Cases._removeMany(case_ids);
			App.Suites.onCasesDeleted();
			App.Suites.refreshGroupCounts();
			App.Cases._disableMassActions();
		},

		error: function error(data) {
			$('#deleteCases').removeClass('button-busy');
			App.Ajax.handleError(data);
		}
	});
};

App.Cases._removeMany = function (case_ids) {
	$.each(case_ids, function (ix, v) {
		$('#row-' + v).remove();
	});
};

App.Cases.onRowClick = function (e) {
	App.Tables.onRowClick(e);
	App.Cases._updateMassActions();
};

App.Cases.onToggleAllClick = function (e) {
	App.Tables.onToggleAllClick(e);
	App.Cases._updateMassActions();
};

App.Cases._updateMassActions = function () {
	var checkboxes = $('tr.row input.selectionCheckbox');

	// Iterate all rows and check if at least one test is selected.
	var checked = false;
	$.each(checkboxes, function (checkboxIndex, checkboxValue) {
		if (checkboxValue.checked) {
			checked = true;
			return;
		}
	});

	if (checked) {
		App.Cases._enableMassActions();
	} else {
		App.Cases._disableMassActions();
	}
};

App.Cases._enableMassActions = function () {
	// Show the enabled mass actions
	$('#deleteCasesDisabled').hide();
	$('#deleteCases').show();
	$('#editCasesSelectedDisabled').hide();
	$('#editCasesSelected').show();
};

App.Cases._disableMassActions = function () {
	// Show the disabled mass actions
	$('#deleteCases').hide();
	$('#deleteCasesDisabled').show();
	$('#editCasesSelected').hide();
	$('#editCasesSelectedDisabled').show();
};

App.Cases.editSelected = function (suite_id) {
	var action = Consts.ajaxBaseUrl + '/cases/edit_many/' + suite_id + // The endpoint
	'/1&' + // Initial load
	App.Suites.getStateOptions(); // The state parameters

	var case_ids = App.Tables.getSelected($('#groups'));
	$('#case_ids').val(case_ids.join(','));
	$('#editCasesForm').attr('action', action).submit();
};

App.Cases.editView = function (suite_id) {
	App.Page.load('/cases/edit_all/{0}/1/{1}&{2}', suite_id, App.Suites.display, App.Suites.getStateOptions() // The state parameters	
	);
};

//---------------------------------------------------------------------
// STEPS
//---------------------------------------------------------------------

App.Cases.stepsToString = function (field_name) {
	var temp = Array();

	$('tr', $('#' + field_name + '_table')).each(function (ix, e) {
		var tr = $(e);

		var o = {
			content: $('textarea.content', tr).val()
		};

		var expected = $('textarea.expected', tr);
		if (expected.length > 0) {
			o.expected = expected.val();
		}

		temp.push(o);
	});

	if (temp.length > 0) {
		return JSON.stringify(temp);
	} else {
		return '';
	}
};

App.Cases.addStep = function (project_id, field_name, step_id) {
	var container = $('#' + field_name + '_container');
	$('.addStepBusy', container).show();
	$('.addStep', container).hide();

	var steps = $('#' + field_name + '_table');
	var index = $('tr', steps).length + 1;

	App.Ajax.call({
		target: '/cases/ajax_render_step',

		arguments: {
			index: index,
			project_id: project_id,
			field_name: field_name
		},

		success: function success(html) {
			$('.noSteps', container).hide();
			var step = $(html);
			if (step_id) {
				var prev = steps.find('tr.step-' + step_id);
				step.insertAfter(prev);
			} else {
				step.appendTo(steps);
			}

			$('.addStepBusy', container).hide();
			$('.addStep', container).show();
			step.find('textarea:first').focus();

			App.Cases.indexSteps(field_name);
			App.Cases.changeSteps(field_name);
		},

		error: function error(data) {
			$('.addStepBusy', container).hide();
			$('.addStep', container).show();
			App.Ajax.handleError(data);
		}
	});
};

App.Cases.addStepToCurrent = function (project_id, field_name) {
	var step_id = App.Cases.getCurrentStep(field_name);
	if (step_id) {
		App.Cases.addStep(project_id, field_name, step_id);
	}
};

App.Cases.updateSteps = function (field_name) {
	var input = $('#' + field_name);
	input.val(App.Cases.stepsToString(field_name));
};

App.Cases.changeSteps = function (field_name) {
	App.Cases.updateSteps(field_name);
	// Not fired otherwise because it's a hidden field
	$('#' + field_name).trigger('change');
	$('#' + field_name).trigger('input');
};

App.Cases.indexSteps = function (field_name) {
	var steps = $('#' + field_name + '_table');

	var current_index = 0;
	$('tr', steps).each(function (ix, e) {
		++current_index;
		$('.step-no-inner', $(e)).html(current_index);
	});
};

App.Cases.removeStep = function (field_name, step_id) {
	var steps = $('#' + field_name + '_table');
	steps.find('tr.step-' + step_id).remove();
	$(step_id).trigger('input');

	if ($('tr', steps).length == 0) {
		$('#' + field_name + '_container').find('.noSteps').show();
	} else {
		App.Cases.indexSteps(field_name);
	}

	App.Cases.changeSteps(field_name);
};

App.Cases.moveStepUp = function (field_name, step_id) {
	var step = $('#' + field_name + '_table').find('tr.step-' + step_id);
	step.insertBefore(step.prevAll('tr').first());
	step.find('textarea').first().focus();
	App.Cases.indexSteps(field_name);
	App.Cases.changeSteps(field_name);
};

App.Cases.moveCurrentStepUp = function (field_name) {
	var step_id = App.Cases.getCurrentStep(field_name);
	if (step_id) {
		App.Cases.moveStepUp(field_name, step_id);
	}
};

App.Cases.getCurrentStep = function (field_name) {
	var focused = $(':focus');
	var step = focused.closest('tr.step-container');
	return step.length > 0 ? step.attr('rel') : null;
};

App.Cases.moveStepDown = function (field_name, step_id) {
	var step = $('#' + field_name + '_table').find('tr.step-' + step_id);
	step.insertAfter(step.nextAll('tr').first());
	step.find('textarea').first().focus();
	App.Cases.indexSteps(field_name);
	App.Cases.changeSteps(field_name);
};

App.Cases.moveCurrentStepDown = function (field_name) {
	var step_id = App.Cases.getCurrentStep(field_name);
	if (step_id) {
		App.Cases.moveStepDown(field_name, step_id);
	}
};

//---------------------------------------------------------------------
// FORM
//---------------------------------------------------------------------

App.Cases.reloadForm = function (suite_id, template_id) {
	$('#templateBusy').show();

	App.Ajax.call({
		target: '/cases/ajax_render_form',

		arguments: $.extend({
			suite_id: suite_id
		}, App.Cases._getFormValues()),

		stop: function stop() {
			$('#templateBusy').hide();
		},

		success: function success(html) {
			$('#form-controls').html(html);
			$('#template_id').trigger('input');
		},

		error: function error(data) {
			App.Ajax.handleError(data);
		}
	});
};

App.Cases.reloadFormMany = function (suite_id, template_id, various) {
	$('#templateBusy').show();

	App.Ajax.call({
		target: '/cases/ajax_render_form_many',

		arguments: $.extend({
			suite_id: suite_id,
			various: various
		}, App.Cases._getFormValues()),

		stop: function stop() {
			$('#templateBusy').hide();
		},

		success: function success(html) {
			$('#form-controls').html(html);
		},

		error: function error(data) {
			App.Ajax.handleError(data);
		}
	});
};

App.Cases._getFormValues = function () {
	// Read the standard input fields from the form.
	var values = {
		title: $('#title').val(),
		section_id: $('#section_id').val(),
		template_id: $('#template_id').val(),
		type_id: $('#type_id').val(),
		priority_id: $('#priority_id').val(),
		estimate: $('#estimate').val(),
		milestone_id: $('#milestone_id').val(),
		refs: $('#refs').val()
	};

	// If there's a step field, make sure to synchronize the steps with
	// the form field.
	var steps = $('#form .custom.steps');
	if (steps.length) {
		App.Cases.updateSteps(steps.attr('id'));
	}

	// And handle and append the custom fields. All custom field
	// controls have a 'custom' class in order to to identify them as
	// custom fields.
	$('#form .custom').each(function (ix, e) {
		var t = $(this);
		if (t.attr('type') == 'checkbox') {
			values[t.attr('name')] = t.prop('checked');
		} else {
			var k = t.attr('name');
			values[k] = t.val();
		}
	});

	// Also inject the included fields for the bulk-edit form, if any.
	// They are not used in single-edit mode.
	$.each(values, function (ix, v) {
		var key = ix + '_included';
		var included = $('#' + key);

		// The included field can either be a checkbox, or a hidden
		// input.
		if (included.is('input[type=checkbox]')) {
			values[key] = included.is(':checked');
		} else {
			values[key] = included.val();
		}
	});

	values['template_id_included'] = true; // Always on after change
	return values;
};

//---------------------------------------------------------------------
// NEXT/PREVIOUS
//---------------------------------------------------------------------

App.Cases.loadPrev = function (case_id, location) {
	App.Ajax.call({
		target: '/cases/ajax_get_prev',

		arguments: {
			case_id: case_id
		},

		success: function success(data) {
			if (data.case_id) {
				document.location = Consts.ajaxBaseUrl + '/cases/' + location + '/' + data.case_id;
			} else {
				$('#directionPrev').hide();
				$('#directionPrevDisabled').show();
			}
		},

		error: function error(data) {
			App.Ajax.handleError(data);
		}
	});
};

App.Cases.loadNext = function (case_id, location) {
	App.Ajax.call({
		target: '/cases/ajax_get_next',

		arguments: {
			case_id: case_id
		},

		success: function success(data) {
			if (data.case_id) {
				document.location = Consts.ajaxBaseUrl + '/cases/' + location + '/' + data.case_id;
			} else {
				$('#directionNext').hide();
				$('#directionNextDisabled').show();
			}
		},

		error: function error(data) {
			App.Ajax.handleError(data);
		}
	});
};

App.Cases.hideStepsHint = function () {
	App.Ajax.call({
		target: '/cases/ajax_hide_steps_hint',
		blockUI: false,

		success: function success(data) {
			App.Effects.hide('#stepsHint');
		},

		error: function error(data) {
			// Non-critical error, can be safely ignored.
		}
	});
};

//---------------------------------------------------------------------
// Dynamic Filters
//---------------------------------------------------------------------

App.Cases.serializeForm = function(data={}) {
	$.each($('#form').serializeArray(), function (i, field) {
		data[field.name] = field.value;
	});

	return data;
}

App.Cases.isDynamicFilters = function(case_id, success_function){
	var form = App.Cases.serializeForm();
	App.Cases.callAjaxCheckCaseInDynamicFilters(case_id, form).done(function(resp) {
		resp = JSON.parse(resp);
		if (resp.is_added) {
			App.Dialogs.confirm(resp.info, success_function, function() {}, '#confirmDialogOkCancel');
		} else {
			success_function();
		}
	}).fail(function(resp) {
		App.Ajax.handleError(resp);
	})
};

App.Cases.callAjaxCheckCaseInDynamicFilters = function(case_id, form) {
	return App.Ajax.call({
		target: '/cases/ajax_check_case_in_run_dynamic',
		arguments: {
			case_id: case_id,
			form: JSON.stringify(form)
		},
	});

};
